
#ifndef TOKEN_EXPORT_H
#define TOKEN_EXPORT_H

#ifdef TOKEN_STATIC_DEFINE
#  define TOKEN_EXPORT
#  define TOKEN_NO_EXPORT
#else
#  ifndef TOKEN_EXPORT
#    ifdef token_EXPORTS
        /* We are building this library */
#      define TOKEN_EXPORT __attribute__((visibility("default")))
#    else
        /* We are using this library */
#      define TOKEN_EXPORT __attribute__((visibility("default")))
#    endif
#  endif

#  ifndef TOKEN_NO_EXPORT
#    define TOKEN_NO_EXPORT __attribute__((visibility("hidden")))
#  endif
#endif

#ifndef TOKEN_DEPRECATED
#  define TOKEN_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef TOKEN_DEPRECATED_EXPORT
#  define TOKEN_DEPRECATED_EXPORT TOKEN_EXPORT TOKEN_DEPRECATED
#endif

#ifndef TOKEN_DEPRECATED_NO_EXPORT
#  define TOKEN_DEPRECATED_NO_EXPORT TOKEN_NO_EXPORT TOKEN_DEPRECATED
#endif

/* NOLINTNEXTLINE(readability-avoid-unconditional-preprocessor-if) */
#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef TOKEN_NO_DEPRECATED
#    define TOKEN_NO_DEPRECATED
#  endif
#endif

#endif /* TOKEN_EXPORT_H */
